module data_output
use precision, only: lk
implicit none

public

integer frcdiag, frcdata1d, partdiag

save
contains
!=================================================================================
subroutine diagnosis
    use global_parameters
    use grids_fields
    use particle_array
    use spline_function,only: spline2d
    use equilibrium, only: B0Mag_Eq, lsr, lsz, spdR, spdZ, zmax
    use read_in_parameters
    use fluid
    use grids_setup
    use grids_eqdata
    implicit none
  
    integer,parameter :: nfield=1,mfdiag=7,modes=1, nfield2=11
    integer :: para_grids
    integer :: i,j,m,ij,frctracking,icount,ierror,nf,nrsp,nzsp, &
               ii, ii_start, ii_end,ip, ipsi_diag
    real(lk) :: fieldtime(mfdiag,nfield),rdum,zdum,sdr,sdz,sdr_inv,sdz_inv,etotdum,bdum,psidum
    real(lk) :: fieldmode(2,modes,nfield),field_flux(mtoroidal),partdata(mpdiag,nspecies)

101 format(i6)
102 format(e13.6)
    frctracking=624

    if(diag_region==1)then

        ii=igrid_core(iflux_core) + min(int(para_diag_location_core*real(mpara_core(iflux_core))) + 1, mpara_core(iflux_core))
        ii_start=igrid_core(iflux_core) + 1
        ii_end=igrid_core(iflux_core) + mpara_core(iflux_core)
        ipsi_diag=iflux_core

    elseif(diag_region==2)then

        ii=mgrid_core + igrid_sol(iflux_sol) + min(int(para_diag_location_sol*real(mpara_sol(iflux_sol))) + 1, mpara_sol(iflux_sol))
        ii_start=mgrid_core+igrid_sol(iflux_sol)+1
        ii_end=mgrid_core+igrid_sol(iflux_sol)+mpara_sol(iflux_sol)
        ipsi_diag=iflux_sol

    else

        write(gtcout,*)"wrong diag_region, diag_region=",diag_region
        stop

    endif

    para_grids = ii_end - ii_start + 1

    if(mype==0 .and. istep==ndiag)call opendiag(nfield2, para_grids)
    call MPI_BCAST(mstepall,1,MPI_INTEGER,0,MPI_COMM_WORLD,ierror)
 
    fieldmode=0.0_lk
    field_flux=0.0_lk

    call spectrum(nfield,modes,phi,fieldmode,field_flux)
    partdata=0.0_lk
    if(ifullk==1)then
      diagfki(7)=-diagfki(7)/(rho0*rho0*kapanip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
      diagfki(9)=-diagfki(9)/(rho0*rho0*kapatip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
      data1dfki(:,1)=-data1dfki(:,1)/(rho0*rho0*kapanip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
      data1dfki(:,3)=-data1dfki(:,3)/(rho0*rho0*kapatip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
    else
      diagion(7)=-diagion(7)/(rho0*rho0*kapanip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
      diagion(9)=-diagion(9)/(rho0*rho0*kapatip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
      data1di(:,1)=-data1di(:,1)/(rho0*rho0*kapanip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
      data1di(:,3)=-data1di(:,3)/(rho0*rho0*kapatip(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshni(ii))
    endif
    if(kinetice==1)then
      diagelectron(7)=-diagelectron(7)/(rho0*rho0*kapanep(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshne(ii))
      diagelectron(9)=-diagelectron(9)/(rho0*rho0*kapatep(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshne(ii))
      data1de(:,1)=-data1de(:,1)/(rho0*rho0*kapanep(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshne(ii))
      data1de(:,3)=-data1de(:,3)/(rho0*rho0*kapatep(ipsi_diag)*Rsol(ii)*bmesh_bsp(ii)*meshne(ii))
    endif
    do ip=1,nspecies
      if(ip==1 .and. ifullk==1) partdata(:,ip)=diagfki/diagfki(11)
      if(ip==1 .and. ifullk==0) partdata(:,ip)=diagion/diagion(11)
      if(ip==2) partdata(:,ip)=diagelectron/diagelectron(11)
    enddo
    if(mype==0)then

        if(istep==1 .and. irk==1)write(gtcout,*)"diagnosis grid index=",ii
        write(frcdiag,102)phi0(1,ii)/meshte(ii)!phi(1,ii)/meshte(ii) ! normalized to e\phi/T_e
        write(frcdiag,102)phi_real(ii)/meshte(ii) ! normalized to e\phi/T_e
        write(frcdiag,102)phi_imag(ii)/meshte(ii) ! normalized to e\phi/T_e
        do i=1,nfield
            do j=1,modes
                write(frcdiag,102)phi0_real(ii)/meshte(ii), phi0_imag(ii)/meshte(ii)!fieldmode(1,j,i)/meshte(ii),fieldmode(2,j,i)/meshte(ii)
            enddo
        enddo
        write(frcdiag,102)densitye(1,ii)
        write(frcdiag,102)densitye0(1,ii)!sdeltne(1,ii)
        if(ifullk==0)then
        write(frcdiag,102)densityi(1,ii)!flowe(1,ii)
        write(frcdiag,102)densityi0(1,ii)
        else
        write(frcdiag,102)densityfki(1,ii)
        write(frcdiag,102)densityfki0(1,ii)
        endif
        write(frcdiag,102)phi_rms/meshte(ii)
        write(frcdiag,102)phi0_rms/meshte(ii)
        call FLUSH(frcdiag)
        do i=1,nspecies
          do j=1,mpdiag
             write(partdiag,102)partdata(j,i)
          enddo
        enddo
        call FLUSH(partdiag)
  
        if(data1d_on==1)then
            if(ifullk==0)then
            write(frcdata1d,102)data1di(:,1)   !particle flux
            write(frcdata1d,102)data1di(:,2)   !mom. flux
            write(frcdata1d,102)data1di(:,3)   !energy flux
            write(frcdata1d,102)data1di(:,4)/meshte(ii)   !phi_rms1d
            write(frcdata1d,102)data1di(:,5)/meshte(ii)   !phi0_rms1d
            else
            write(frcdata1d,102)data1dfki(:,1)   !particle flux
            write(frcdata1d,102)data1dfki(:,2)   !mom. flux
            write(frcdata1d,102)data1dfki(:,3)   !energy flux
            write(frcdata1d,102)data1dfki(:,4)/meshte(ii)   !phi_rms1d
            write(frcdata1d,102)data1dfki(:,5)/meshte(ii)   !phi0_rms1d
            endif
            if(kinetice==1)then
            write(frcdata1d,102)data1de(:,1)   !particle flux
            write(frcdata1d,102)data1de(:,2)   !mom. flux
            write(frcdata1d,102)data1de(:,3)   !energy flux
            endif
            call FLUSH(frcdata1d)
        endif
        write(gtcout,"(I7,6e15.6)")istep+mstepall,phi(1,ii)/meshte(ii)
        call FLUSH(gtcout)
    endif

    if(.false.)then
    if(mype==0 .and. istep==ndiag .and. irun==0)then
  ! open time history file in new run
        open(frctracking,file='frc_tracking.out',status='replace')
    endif
  
    if(mype==0)then
        rdum=zelectron(1,1)
        zdum=zelectron(3,1)
        bdum = spline2d(rdum,zdum+zmax,0,lsr,lsz,spdR,spdZ,B0Mag_Eq,0,0)
        write(frctracking,102)istep,zelectron(1,1),zelectron(2,1),zelectron(3,1),zelectron(6,1),zelectron(8,1),zelectron(7,1),bdum
        call FLUSH(frctracking)
    endif
    endif
end subroutine diagnosis
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine opendiag(nfield2, para_grids)
    use global_parameters
    use read_in_parameters, only: ndiag, tstep, mstep, irun, ifullk, kinetice
    use particle_array, only: mpdiag
    use grids_setup, only: mpsi
    implicit none

    integer, intent(in) :: nfield2, para_grids
    integer :: i, j, ndstep, iotmp, ndata, idum, num_var1,num_var2
    real(lk) :: time_step_tmp, tdum
    character(len=100) cdum
    character(len=25) :: txt25
    character(len=99) :: tmpline
    integer :: countline1,num_Var,ndim(6),tmpid,tmpSize
 

    frcdiag=654
    partdiag=557
    frcdata1d=987
    iotmp=321
   
    ndstep=mstep/ndiag
    ndata=nfield2
    if(irun==0)then
        mstepall=0
        open(frcdiag, file='frc_history.out', status='replace')

        write(frcdiag,101)ndstep
        write(frcdiag,102)tstep*real(ndiag)
        if(data1d_on==1)then
            open(frcdata1d, file='frc_data1d.out',status='replace')
            num_var1=3
            write(frcdata1d,'(A25,i6)')'#onetime_variable_num',num_var1
            write(frcdata1d,'(A25,i6,i6)')'#mpsi',1,1
            write(frcdata1d,'(A25,i6,i6)')'#dt',1,1
            write(frcdata1d,'(A25,i6,i6)')'#ndstep',1,1

            num_var2=5
            if(kinetice==1)then
            num_var2=num_var2+3
            endif
            write(frcdata1d,'(A25,i6)')'#data1d_output_num',num_var2
            if(ifullk==0)then
            write(frcdata1d,'(A25,i6,i6)')'#partflux1di',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#momflux1di',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#heatflux1di',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#phi_rms1d',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#phi0_rms1d',1,mpsi
            else
            write(frcdata1d,'(A25,i6,i6)')'#partflux1dfki',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#momflux1dfki',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#heatflux1dfki',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#phi_rms1d',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#phi0_rms1d',1,mpsi
            endif
            if(kinetice==1)then
            write(frcdata1d,'(A25,i6,i6)')'#partflux1de',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#momflux1de',1,mpsi
            write(frcdata1d,'(A25,i6,i6)')'#heatflux1de',1,mpsi
            endif

            write(frcdata1d,101)mpsi
            write(frcdata1d,102)tstep*real(ndiag)
            write(frcdata1d,101)ndstep
            call flush(frcdata1d)
        endif
        open(partdiag, file='partdiag.out',status='replace')
        write(partdiag,101)mpdiag
        write(partdiag,102)tstep*real(ndiag)
        write(partdiag,101)nspecies
    else
! find history file from previous run
        irest = irest - 1
        if(mod(irest,2)==0)then
            cdum="restart_dir1/frc_history_restart.out"   
        else
            cdum="restart_dir2/frc_history_restart.out"
        endif
!=================
        open(iotmp, file=trim(cdum),status='old')
        rewind(iotmp)
!=================
! open time history file for restart run
        open(frcdiag,file='frc_history.out',status='replace')

! # of time steps
        read(iotmp,101)mstepall
        write(frcdiag,101)mstepall+ndstep
 
        read(iotmp,102)time_step_tmp
        write(frcdiag,102)time_step_tmp

        do i=1, ndata*mstepall
            read(iotmp,102)tdum
            write(frcdiag,102)tdum           
        enddo  
        close(iotmp)

        if(data1d_on==1)then
! copy restart data1d.out file
            open(frcdata1d,file='frc_data1d.out',status='replace')
            if(mod(irest,2)==0)then
                cdum="restart_dir1/frc_data1d_restart.out"
            else
                cdum="restart_dir2/frc_data1d_restart.out"
            endif
    
            open(iotmp,file=trim(cdum),status='old')
            rewind(iotmp)

            countline1=0
            ndata=0

            read(iotmp,'(A,i6)')txt25,num_var
            write(frcdata1d,'(A,i6)')txt25,num_var
            do i=1,num_var
                read(iotmp,'(A)')tmpline
                txt25=tmpline(1:25)
                read(tmpline(26:31),'(i6)') ndim(1)
                tmpSize = 1
                do j=1,ndim(1)
                    tmpid = 26+j*6
                    read(tmpline(tmpid:tmpid+6),*) ndim(j+1)
                    tmpSize=tmpSize*ndim(j+1)
                enddo
                countline1=countline1+1
                write(frcdata1d,'(A)')trim(tmpline)
            enddo

            read(iotmp,'(A,i6)')txt25,num_var
            write(frcdata1d,'(A,i6)')txt25,num_var
            do i=1,num_var
                read(iotmp,'(A)')tmpline
                txt25=tmpline(1:25)
                read(tmpline(26:31),'(i6)') ndim(1)
                tmpSize = 1
                do j=1,ndim(1)
                    tmpid = 26+j*6
                    read(tmpline(tmpid:tmpid+6),*) ndim(j+1)
                    tmpSize=tmpSize*ndim(j+1)
                enddo
                ndata=ndata+tmpSize
                write(frcdata1d,'(A)')trim(tmpline)
            enddo

            do i=2,countline1
                read(iotmp,'(A)')tmpline
                write(frcdata1d,'(A)')trim(tmpline)
            enddo
! # of time steps        
            read(iotmp,101)mstepall
            write(frcdata1d,101)mstepall+ndstep

            do i=1,ndata*mstepall
                read(iotmp,'(A)')tmpline
                write(frcdata1d,'(A)')trim(tmpline)
            enddo
            close(iotmp)
        endif

        open(partdiag, file='partdiag.out',status='old',access='append') 
! restore irest value
        irest = irest + 1
!!calculate current total time step
        mstepall=mstepall*ndiag

    endif

101 format(i6)
102 format(e13.6)

end subroutine opendiag
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine snapshot
    use precision
    use global_parameters
    use simulation_domain, only: sfsol, sfcore
    use read_in_parameters
    use grids_setup, only: mpsi_core, mpara_core, igrid_core, mgrid_core, &
                           mpsi_sol, mpara_sol, igrid_sol
    use grids_fields, only: phi, phi_r_zeta_core, phi_r_zeta_sol

! for pressure output
    use grids_setup, only: mpsi
    use particle_array, only: marknumi,marknume,markeri,markere,pressureipara,pressureiperp, &
                              pressureepara, pressureeperp
    use grids_setup, only: igrid_core,igrid_sol
    use simulation_domain, only: sfcore,sfsol
   
    implicit none
  
    integer :: i,j,ij,icount,ierror
    real(lk),dimension(:),allocatable :: eachflux_core, eachflux_sol
    real(lk),dimension(:,:),allocatable :: allflux_core, allflux_sol

! for pressure output
    real(lk) :: sumpipara(mpsi),sumpiperp(mpsi),sumpepara(mpsi),sumpeperp(mpsi), &
                summarkeri(mpsi),summarkere(mpsi),sumdum(mpsi)
 
    if(sfcore==1)then
        allocate(eachflux_core(mpsi_core), allflux_core(mpsi_core,mtoroidal))
        do i=1,mpsi_core
            ij=igrid_core(i)+1
            eachflux_core(i)=phi(1,ij)
        enddo
        icount=mpsi_core
        call MPI_GATHER(eachflux_core,icount,mpi_Rsize,allflux_core,icount,mpi_Rsize,0,toroidal_comm,ierror)
        do i=1,mpsi_core
            do j=1,mtoroidal
                ij=(i-1)*mtoroidal+j
                phi_r_zeta_core(ij) = allflux_core(i,j)
            enddo
        enddo
        deallocate(eachflux_core,allflux_core)
    endif

    if(sfsol==1)then
        allocate(eachflux_sol(mpsi_sol),allflux_sol(mpsi_sol,mtoroidal))
        do i=1,mpsi_sol
            ij=mgrid_core+igrid_sol(i)+min(int(para_diag_location_sol*real(mpara_sol(i))) + 1, mpara_sol(i))
            eachflux_sol(i)=phi(1,ij)
        enddo
        icount=mpsi_sol
        call MPI_GATHER(eachflux_sol,icount,mpi_Rsize,allflux_sol,icount,mpi_Rsize,0,toroidal_comm,ierror)
        do i=1,mpsi_sol
            do j=1,mtoroidal
                ij=(i-1)*mtoroidal+j
                phi_r_zeta_sol(ij) = allflux_sol(i,j)
            enddo
        enddo
        deallocate(eachflux_sol,allflux_sol)
    endif


!compute temperature in \psi
if(ifullk==0)then
    sumpipara=0.0_lk
    sumpiperp=0.0_lk
    sumpepara=0.0_lk
    sumpeperp=0.0_lk
    summarkeri=0.0_lk
    summarkere=0.0_lk
    if(sfcore==1)then
        do i=1,mpsi_core
            do j=1,mpara_core(i)
                ij=igrid_core(i)+j
                sumpipara(i)=sumpipara(i)+pressureipara(1,ij)/markeri(ij)
                sumpiperp(i)=sumpiperp(i)+pressureiperp(1,ij)/markeri(ij)
                summarkeri(i)=summarkeri(i)+1.0_lk/markeri(ij)

                sumpepara(i)=sumpepara(i)+pressureepara(1,ij)/markere(ij)
                sumpeperp(i)=sumpeperp(i)+pressureeperp(1,ij)/markere(ij)
                summarkere(i)=summarkere(i)+1.0_lk/markere(ij)
            enddo
        enddo
    endif
    if(sfsol==1)then
        do i=1,mpsi_sol
            do j=1,mpara_sol(i)
                ij=mgrid_core+igrid_sol(i)+j
                sumpipara(i+mpsi_core)=sumpipara(i+mpsi_core)+pressureipara(1,ij)/markeri(ij)
                sumpiperp(i+mpsi_core)=sumpiperp(i+mpsi_core)+pressureiperp(1,ij)/markeri(ij)
                summarkeri(i+mpsi_core)=summarkeri(i+mpsi_core)+1.0_lk/markeri(ij)

                sumpepara(i+mpsi_core)=sumpepara(i+mpsi_core)+pressureepara(1,ij)/markere(ij)
                sumpeperp(i+mpsi_core)=sumpeperp(i+mpsi_core)+pressureeperp(1,ij)/markere(ij)
                summarkere(i+mpsi_core)=summarkere(i+mpsi_core)+1.0_lk/markere(ij)
            enddo
        enddo
    endif
    sumdum=0.0_lk
    call MPI_ALLREDUCE(sumpipara,sumdum,mpsi,mpi_Rsize,MPI_SUM,toroidal_comm,ierror)
    sumpipara=sumdum
    call MPI_ALLREDUCE(sumpiperp,sumdum,mpsi,mpi_Rsize,MPI_SUM,toroidal_comm,ierror)
    sumpiperp=sumdum
    call MPI_ALLREDUCE(sumpepara,sumdum,mpsi,mpi_Rsize,MPI_SUM,toroidal_comm,ierror)
    sumpepara=sumdum
    call MPI_ALLREDUCE(sumpeperp,sumdum,mpsi,mpi_Rsize,MPI_SUM,toroidal_comm,ierror)
    sumpeperp=sumdum
    call MPI_ALLREDUCE(summarkeri,sumdum,mpsi,mpi_Rsize,MPI_SUM,toroidal_comm,ierror)
    summarkeri=sumdum
    call MPI_ALLREDUCE(summarkere,sumdum,mpsi,mpi_Rsize,MPI_SUM,toroidal_comm,ierror)
    summarkere=sumdum
    
    if(sfcore==1)then
        do i=1,mpsi_core
            do j=1,mpara_core(i)
                ij=igrid_core(i)+j
                pressureipara(1,ij)=sumpipara(i)/summarkeri(i)/1.5_lk
                pressureiperp(1,ij)=sumpiperp(i)/summarkeri(i)/1.5_lk

                pressureepara(1,ij)=sumpepara(i)/summarkere(i)/1.5_lk
                pressureeperp(1,ij)=sumpeperp(i)/summarkere(i)/1.5_lk
             enddo
        enddo
    endif
    if(sfsol==1)then
        do i=1,mpsi_sol
            do j=1,mpara_sol(i)
                ij=igrid_sol(i)+j
                pressureipara(1,ij)=sumpipara(i+mpsi_core)/summarkeri(i+mpsi_core)/1.5_lk
                pressureiperp(1,ij)=sumpiperp(i+mpsi_core)/summarkeri(i+mpsi_core)/1.5_lk

                pressureepara(1,ij)=sumpepara(i+mpsi_core)/summarkere(i+mpsi_core)/1.5_lk
                pressureeperp(1,ij)=sumpeperp(i+mpsi_core)/summarkere(i+mpsi_core)/1.5_lk
            enddo
        enddo 
    endif
endif


    if(mype==0)then
        call perturbed_field_mype0
!        call fullf_mype0
!        call deltaf_snap
    endif

    call MPI_BARRIER(MPI_COMM_WORLD,ierror)
end subroutine snapshot
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine fullf_mype0
    use precision
    use particle_array
    use grids_setup
    use equilibrium
    use mapping_splines
    use simulation_domain
    use read_in_parameters
    use global_parameters
    use spline_function, only: spline2d
    use cylindrical_function, only: geotheta
    implicit none

    integer, parameter :: ntgrid=40
    real(lk) :: pdf(mpsi,ntgrid,2), fullf, pdum, sdum, tdum, delp_core, delp_sol, h
    integer :: ip, it, m, frcfullffile, nsnap
    character(len=99) snapshot_fname, fname, format_string
    101 format(i6)
    102 format(e13.6)

    frcfullffile=325 !open frc equilibrium plot output file
    format_string = "(A16,I0.5)" ! 7 characters for "restart"
                                         ! 5 characters for mpi id with padded zeros
    nsnap=mstepall+istep
    write(fname,'(i7.7,".out")')nsnap
    snapshot_fname='fullf'//trim(fname)
    open(frcfullffile,file=snapshot_fname,status='replace')

    pdf = 0.0_lk
    if(sfcore==1) delp_core = (psi1_core-psi0_core)/(mpsi_core-1)
    if(sfsol==1) delp_sol = (psi1_sol-psi0_sol)/(mpsi_sol-1)
    if(ifullk==0)then
    do m=1,mi
       pdum = spline2d(zion(1,m),zion(3,m)+zmax,0,lsr,lsz,spdR,spdZ,psirz_Eq,0,0)
       if(pdum<psi1_core)then
       ip = modulo(pdum,delp_core)
       h = (pdum-ip*delp_core)/delp_core
       tdum=geotheta(zion(1,m),zion(3,m))
       sdum=spline2d(pdum-msp_psi_core(1),tdum,0,lspcore,lstcore,del_psi_core,del_theta_core,sspcore_psi_theta,0,0)
       else
       ip = modulo(pdum-psi1_core,delp_sol)
       h = (pdum-psi1_core-ip*delp_sol)/delp_sol
       ip = ip + mpsi_core
       sdum=spline2d(pdum-msp_psi_sol(1),zion(3,m)-Zsol_domain0,0,lspsol,lszsol,del_psi_sol,del_z_sol,sspsol_psi_z,0,0)
       endif
       it = int(ceiling(1.0*sdum * ntgrid))
       fullf = zion0(6,m)
       pdf(ip+1,it,1)=pdf(ip+1, it, 1) + fullf*(1.0_lk-h)
       pdf(ip+2,it,1)=pdf(ip+2,it,1)+fullf*h
    enddo
    else
    do m=1,mfki
       pdum = spline2d(zfki(1,m),zfki(3,m)+zmax,0,lsr,lsz,spdR,spdZ,psirz_Eq,0,0)
       if(pdum<psi1_core)then
       ip = modulo(pdum,delp_core)
       h = (pdum-ip*delp_core)/delp_core
       tdum=geotheta(zfki(1,m),zfki(3,m))
       sdum=spline2d(pdum-msp_psi_core(1),tdum,0,lspcore,lstcore,del_psi_core,del_theta_core,sspcore_psi_theta,0,0)
       else
       ip = modulo(pdum-psi1_core,delp_sol)
       h = (pdum-psi1_core-ip*delp_sol)/delp_sol
       ip = ip + mpsi_core
       sdum=spline2d(pdum-msp_psi_sol(1),zfki(3,m)-Zsol_domain0,0,lspsol,lszsol,del_psi_sol,del_z_sol,sspsol_psi_z,0,0)
       endif
       it = ceiling(sdum * ntgrid)
       fullf = zfki0(7,m)
       pdf(ip+1,it,1)=pdf(ip+1, it, 1) + fullf*(1.0_lk-h)
       pdf(ip+2,it,1)=pdf(ip+2,it,1)+fullf*h
    enddo
    endif
    if(kinetice==1)then
    do m=1,me
       pdum = spline2d(zelectron(1,m),zelectron(3,m)+zmax,0,lsr,lsz,spdR,spdZ,psirz_Eq,0,0)
       if(pdum<psi1_core)then
       ip = modulo(pdum,delp_core)
       h = (pdum-ip*delp_core)/delp_core
       tdum=geotheta(zelectron(1,m),zelectron(3,m))
       sdum=spline2d(pdum-msp_psi_core(1),tdum,0,lspcore,lstcore,del_psi_core,del_theta_core,sspcore_psi_theta,0,0)
       else
       ip = modulo(pdum-psi1_core,delp_sol)
       h = (pdum-psi1_core-ip*delp_sol)/delp_sol
       ip = ip + mpsi_core
       sdum=spline2d(pdum-msp_psi_sol(1),zelectron(3,m)-Zsol_domain0,0,lspsol,lszsol,del_psi_sol,del_z_sol,sspsol_psi_z,0,0)
       endif
       it = ceiling(sdum * ntgrid)
       fullf = zelectron0(6,m)
       pdf(ip+1,it,2)=pdf(ip+1, it, 2) + fullf * (1.0_lk-h)
       pdf(ip+2,it,1)=pdf(ip+2,it,1)+fullf*h
    enddo
    endif
    write(frcfullffile,102) pdf
    close(frcfullffile)

end subroutine fullf_mype0
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine perturbed_field_mype0
    use precision
    use global_parameters, only: istep,mstepall, mtoroidal,mype
    use grids_fields, only: phi, phi_real, phi_imag, phi_r_zeta_core, phi_r_zeta_sol,&
                            phi0, phi0_real, phi0_imag
    use particle_array,only: densityi, flowi, pressureipara, pressureiperp, &
                             densitye, flowe, pressureepara, pressureeperp, &
                             densityfki, flowfki, pressurefkipara, pressurefkiperp, &
                             densityi0, densitye0, data1di
    use read_in_parameters, only: numerfrc,ifullk
    use grids_setup, only: mpsi_core, mpara_core, mgrid_core, &
                           mpsi_sol, mpara_sol, mgrid_sol
    use grids_eqdata, only: meshte   
    implicit none
  
    integer :: i,j,ij,frcsnap,icount,ierror,nsnap
    character(len=99) :: fname,format_string,snapshot_fname
    101 format(i6)
    102 format(e13.6)

    frcsnap=321 !open frc equilibrium plot output file
    format_string = "(A16,I0.5)" ! 7 characters for "restart"
                                         ! 5 characters for mpi id with padded zeros
    nsnap=mstepall+istep
    write(fname,'(i7.7,".out")')nsnap
    snapshot_fname='frc_CYL_snapshot'//trim(fname)
    open(frcsnap,file=snapshot_fname,status='replace')

    if(numerfrc==1)then 
        write(frcsnap,101)mpsi_core,mpara_core(1),mgrid_core
        write(frcsnap,101)mpsi_sol,mpara_sol(1),mgrid_sol
        write(frcsnap,101)mtoroidal   
        write(frcsnap,102)phi(1,:)/meshte,phi_real/meshte,phi_imag/meshte,phi_r_zeta_core/meshte,phi_r_zeta_sol/meshte
        if(ifullk==0)then
        write(frcsnap,102)densityi(1,:),flowi(1,:),pressureipara(1,:),pressureiperp(1,:)
        else
        write(frcsnap,102)densityfki(1,:),flowfki(1,:),pressurefkipara(1,:),pressurefkiperp(1,:)
        endif
        write(frcsnap,102)densitye(1,:),flowe(1,:),pressureepara(1,:),pressureeperp(1,:)

    elseif(numerfrc==2)then
        write(frcsnap,101)mpsi_core,mpara_core(1),mgrid_core
        write(frcsnap,101)mtoroidal
        write(frcsnap,102)phi(1,:)/meshte,phi_real/meshte,phi_imag/meshte,phi_r_zeta_core/meshte
        if(ifullk==0)then
        write(frcsnap,102)densityi(1,:),flowi(1,:),pressureipara(1,:),pressureiperp(1,:)
        else
        write(frcsnap,102)densityfki(1,:),flowfki(1,:),pressurefkipara(1,:),pressurefkiperp(1,:)
        endif
        write(frcsnap,102)densitye(1,:),flowe(1,:),pressureepara(1,:),pressureeperp(1,:)

    elseif(numerfrc==3 .or. numerfrc==4)then
        write(frcsnap,101)mpsi_sol,mpara_sol(1),mgrid_sol
        write(frcsnap,101)mtoroidal
        write(frcsnap,102)phi(1,:)/meshte,phi_real/meshte,phi_imag/meshte,phi_r_zeta_sol
!        write(frcsnap,102)phi0(1,:)/meshte,phi0_real/meshte,phi0_imag/meshte,phi_r_zeta_sol
        if(ifullk==0)then
        !write(frcsnap,102)densityi(1,:),flowi(1,:),pressureipara(1,:),pressureiperp(1,:)
        write(frcsnap,102)densityi(1,:),densityi0(1,:),phi0(1,:)/meshte,pressureiperp(1,:)+pressureipara(1,:)
        else
        write(frcsnap,102)densityfki(1,:),flowfki(1,:),pressurefkipara(1,:),pressurefkiperp(1,:)
        endif
        !write(frcsnap,102)densitye(1,:),flowe(1,:),pressureepara(1,:),pressureeperp(1,:)
        write(frcsnap,102)densitye(1,:),densitye0(1,:),phi0_real/meshte,phi0_imag/meshte
    endif
    flush(frcsnap)

! output radial dtem and dn information
    write(fname,'(i7.7,".out")')mstepall+istep
    fname='dtem_output'//trim(fname)
    open(404,file=fname,status='replace')
    write(404,102)data1di(:,6)
    write(404,102)data1di(:,7)
    write(404,102)data1di(:,8)
    write(404,102)data1di(:,9)
    call FLUSH(404)
    close(404)

end subroutine perturbed_field_mype0

subroutine spectrum(nfield,modes,fieldgrid,fieldmode,field_flux)
    use global_parameters
    use grids_setup
    use grids_fields
    use fluid
    use simulation_domain, only: sfcore, sfsol
    use read_in_parameters, only: n_modes, diag_region, iflux_core, iflux_sol, &
                                  para_diag_location_core, para_diag_location_sol
    
    integer,intent(in) :: modes,nfield
    integer i,j,ij,icount,ierror
    real(lk),dimension(0:1,mgrid),intent(in) :: fieldgrid
    real(lk),dimension(2,modes,nfield),intent(out) :: fieldmode
    real(lk),dimension(mtoroidal),intent(out) :: field_flux
    
    real(lk) :: xz(mtoroidal),phifluxlz(mgrid,mtoroidal),&
                eachzeta(mgrid)
    complex(lk) yz(mtoroidal/2+1)
  
! periodic BC in poloidal and toroidal
  
    eachzeta=0.0_lk
    icount=mgrid
    if(sfcore==1)then
!$omp parallel do private(i,j,ij) 
        do i=1,mpsi_core
            do j=1,mpara_core(i)
                ij=igrid_core(i)+j
                eachzeta(ij)=fieldgrid(1,ij)
            enddo
        enddo
    endif
    if(sfsol==1)then
!$omp parallel do private(i,j,ij) 
         do i=1,mpsi_sol
             do j=1,mpara_sol(i)
                 ij=mgrid_core+igrid_sol(i)+j
                 eachzeta(ij)=fieldgrid(1,ij)
             enddo
         enddo
    endif
   
    call MPI_ALLGATHER(eachzeta,icount,mpi_Rsize,phifluxlz,icount,&
            mpi_Rsize,toroidal_comm,ierror)

    if(mype==0)then
        if(diag_region==1)then  
            yz=0.0_lk
            ij=igrid_core(iflux_core) + min(int(para_diag_location_core*real(mpara_core(iflux_core))) + 1, mpara_core(iflux_core))
            xz=phifluxlz(ij,:)
            call fftr1d(1,mtoroidal,1.0_lk,xz,yz,2)
            field_flux=xz
! for exp(-i\omega*t+in*\zeta)
            do i=1,nfield
                do j=1,modes
                    fieldmode(1,j,i)=real(yz(n_modes(j)+1))
                    fieldmode(2,j,i)=-aimag(yz(n_modes(j)+1))
                enddo
            enddo
        elseif(diag_region==2)then
            yz=0.0_lk
            ij=mgrid_core+igrid_sol(iflux_sol)+ min(int(para_diag_location_sol*real(mpara_sol(iflux_sol))) + 1, mpara_sol(iflux_sol))
            xz=phifluxlz(ij,:)
            call fftr1d(1,mtoroidal,1.0_lk,xz,yz,2)
            field_flux=xz
! for exp(-i\omega*t+in*\zeta)
            do i=1,nfield
                do j=1,modes
                    fieldmode(1,j,i)=real(yz(n_modes(j)+1))
                    fieldmode(2,j,i)=-aimag(yz(n_modes(j)+1))
                enddo
            enddo
        endif
    endif
end subroutine spectrum
subroutine deltaf_snap
    use particle_array
    use global_parameters
    use read_in_parameters
    implicit none
    real(lk) :: dfi(100,100), dfe(100,100), dftmp(100,100)
    integer ierror, nsnap
    character(len=99) :: fname, deltaf_fname
    dfi = 0.0_lk
    dfe = 0.0_lk
    dftmp = 0.0_lk
if(ifullk==0)then
    call deltaf_kernel(zion,mi,aion, dfi)
!else
!    call deltaf_kernel(zfki,mfki, dfi)

    call mpi_gather(dfi, 10000, MPI_DOUBLE_PRECISION, dftmp, 10000,&
        MPI_DOUBLE_PRECISION, 0, mpi_comm_world, ierror)
    dftmp = dftmp / numberpe
    dfi = dftmp
endif
    if(kinetice==0)then
        call deltaf_kernel(zelectron,me,aelectron,dfe)

    call mpi_gather(dfe, 10000, MPI_DOUBLE_PRECISION, dftmp, 10000,&
        MPI_DOUBLE_PRECISION, 0, mpi_comm_world, ierror)
    dftmp = dftmp / numberpe
    dfe = dftmp
    endif
    nsnap=mstepall+istep
    write(fname,'(i7.7,".out")')nsnap
    deltaf_fname='deltaf'//trim(fname)
    if(mype==0)then
      open(322,file=deltaf_fname,status='replace')
      write(322,*)ifullk
      write(322,*)kinetice
      write(322,102) dfi
      if(kinetice==1) write(322,102) dfe
      close(322)
    endif

102 format(e13.6)

end subroutine deltaf_snap

subroutine deltaf_kernel(zpart, mp, apart, df)
    use grids_setup
    use grids_eqdata
    use equilibrium
    use global_parameters
    use read_in_parameters
    use spline_function, only: spline2d

    implicit none
    real(lk), dimension(:,:) :: zpart, df
    real(lk) :: apart
    integer mp
    integer, parameter :: mEgrid=100, mlgrid=100
    integer :: iEgrid, ilgrid,m,nrsp,nzsp
    real(lk) :: Emax, lmax, dE, dl,dpr,dr2,dpz,dz2,rdum_inv,gradpsi_r,gradpsi_z,br,bz,bt,b,b_inv,vperp2,&
                vpara2, energy, psidum,lambda,spdR_inv,spdZ_inv
    

    df = 0.0_lk
    Emax = rho0**2*10
    dE = Emax / mEgrid
    lmax = 3
    dl = lmax / mlgrid
    spdR_inv=1.0_lk/spdR
    spdZ_inv=1.0_lk/spdZ
    !determin B
    do m=1,mp
        nrsp=max(1,min(lsr-1,int(ceiling(zpart(1,m)*spdR_inv))))
        dpr=zpart(1,m)-spdR*real(nrsp-1)
        dr2=dpr*dpr
        nzsp=max(1,min(lsz-1,int(ceiling((zpart(3,m)+zmax)*spdZ_inv))))
        dpz=zpart(3,m)+zmax-spdZ*real(nzsp-1)
        dz2=dpz*dpz
        rdum_inv=1.0_lk/zpart(1,m)
        gradpsi_r= psirz_Eq(2,nrsp,nzsp) + 2.0_lk*psirz_Eq(3,nrsp,nzsp)*dpr &
            + psirz_Eq(5,nrsp,nzsp)*dpz + 2.0_lk*psirz_Eq(6,nrsp,nzsp)*dpr*dpz&
            + psirz_Eq(8,nrsp,nzsp)*dz2 + 2.0_lk*psirz_Eq(9,nrsp,nzsp)*dpr*dz2

        gradpsi_z=psirz_Eq(4,nrsp,nzsp) + psirz_Eq(5,nrsp,nzsp)*dpr +psirz_Eq(6,nrsp,nzsp)*dr2 &
                + 2.0_lk*psirz_Eq(7,nrsp,nzsp)*dpz + 2.0_lk*psirz_Eq(8,nrsp,nzsp)*dpr*dpz + 2.0_lk*psirz_Eq(9,nrsp,nzsp)*dr2*dpz
        br=-gradpsi_z*rdum_inv
        bz=gradpsi_r*rdum_inv
        bt=0.0_lk
        b=sqrt(br**2+bz**2)
        b_inv=1.0_lk/b
        vperp2 = zpart(6,m)*zpart(6,m)*b
        vpara2 = 0.5_lk*apart*zpart(4,m)*zpart(4,m)
        energy = vperp2 + vpara2
        !psidum = spline2d(zpart(1,m),zpart(3,m)+zmax,0,lsr,lsz,spdR,spdZ,psirz_Eq,0,0)
        !R0 = spline2d(psidum-msp_psi_sol(1),zmax,0,lspsol,lszsol,del_psi_sol,del_s_sol,rspsol_psi_s,0,0)
        !b0 = spline2d(Rcore(ij),Zcore(ij)+zmax,0,lsr,lsz,spdR,spdZ,B0Mag_Eq,0,0)
        lambda = zpart(6,m)*zpart(6,m)*b0/energy
        iEgrid = min(int(modulo(energy, dE)),mEgrid)
        ilgrid = min(int(modulo(lambda, dl)),mlgrid)
        df(iEgrid, ilgrid) = df(iEgrid, ilgrid) + zpart(5,m)**2
    enddo
end subroutine deltaf_kernel

end module data_output
